import pydantic

from ecgai_drawing.models.model_base import MyBaseModel


class EcgPlotImage(MyBaseModel):
    transaction_id: str
    record_name: str
    file_name: str
    file_extension: str
    image: bytes

    # def __int__(
    #     self, transaction_id: str, record_name: str, file_name: str, image: bytes
    # ):
    #     self.transaction_id = transaction_id
    #     self.record_name = record_name
    #     self.file_name = file_name
    #     self.image = image

    @classmethod
    def create(cls, transaction_id: str, record_name: str, file_name: str, file_extension: str, image: bytes):
        try:
            d = dict(
                TransactionId=transaction_id,
                RecordName=record_name,
                FileName=file_name,
                FileExtension=file_extension,
                Image=image,
            )
            return cls.from_dict(d)
        except pydantic.ValidationError as e:
            # logging.error(e)
            raise e



import pathlib
from math import ceil

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Subplot
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from matplotlib.ticker import AutoMinorLocator
from numpy import ndarray
from pydantic.dataclasses import dataclass

# from definitions import ROOT_DIR
from ecgai_drawing import images
from ecgai_drawing.enums.color_style import ColorStyle
from ecgai_drawing.images import combine_images, load_image
from ecgai_drawing.models.ecg_leads import Leads

# from skimage.COLOR import rgba2rgb


@dataclass
class EcgPlotter:
    class Config:
        use_enum_values = True

    lead_index_order = [
        "I",
        "II",
        "III",
        "aVR",
        "aVL",
        "aVF",
        "V1",
        "V2",
        "V3",
        "V4",
        "V5",
        "V6",
    ]
    lead_order: list = None
    sample_rate: int = 500
    title: str = "ECG 12 lead"
    color_style: ColorStyle = ColorStyle.COLOR
    line_width: float = 1.0
    columns: int = 2
    row_height: int = 6
    show_lead_name: bool = True
    show_grid: bool = False
    show_separate_line: bool = True
    adjust_subplots: bool = True

    def plot(
        self,
        sample_rate: int,
        ecg_leads: Leads,
        title: str = "ECG 12 lead",
        color_style: ColorStyle = ColorStyle.COLOR,
        show_grid: bool = False,
    ) -> ndarray:
        """

        Args:
            title: Title shown at the top of the ECG printout
            sample_rate: Number of signal samples per second
            ecg_leads: Leads collection for drawing onto ECG plot
            color_style: Color style of output, default is COLOR.
            Other options are black and white, MASK and binary format
            show_grid: Show ECG grids on plot

        Returns:
            ndarray:

        """
        self.sample_rate = sample_rate
        self.title = title
        self._set_style(color_style, show_grid)
        image = self._create_plot(ecg_leads)
        # convert from rgba format to rgb format
        image = images.convert_from_rgba_to_rgb(image=image)
        # convert from rgb format to bgr format required by openCv
        image = images.convert_from_rgb_to_bgr(image=image)

        if self.show_grid:
            # consider using different grid image based on the configuration of the Ecg plot
            base_image = load_image(image_name="colour_six_x_two.png", image_path=data_directory())
        else:
            base_image = np.zeros([768, 2048, 3], dtype=np.uint8)
            base_image.fill(255)
            # if self.color_style == ColorStyle.COLOR:
        image = combine_images(base_image=base_image, top_image=image)

        return image

    def _draw_plot(self, ecg_leads: Leads):
        leads = self._convert_to_ndarray(ecg_leads=ecg_leads)

        if not self.lead_order:
            self.lead_order = list(range(len(leads)))

        length_seconds = len(leads[0]) / self.sample_rate
        number_of_leads = len(self.lead_order)
        rows = int(ceil(number_of_leads / self.columns))
        # display_factor = 2.5
        display_factor = 1
        # line_width = 0.5

        figure = self._setup_figure(display_factor, rows, length_seconds)

        x_max, x_min, y_max, y_min = self._set_xy_axis_size(length_seconds, rows)

        subplot = self._setup_subplot(figure, x_max, x_min, y_max, y_min)

        display_factor **= 0.5

        color_line, color_major_grid, color_minor_grid = self._set_line_style()

        if self.show_grid:
            pass
            # self._draw_grid(
            #     subplot=subplot,
            #     color_major_grid=color_major_grid,
            #     color_minor_grid=color_minor_grid,
            #     display_factor=display_factor,
            #     x_max=x_max,
            #     x_min=x_min,
            #     y_max=y_max,
            #     y_min=y_min,
            # )

        self._draw_plot_leads(
            subplot=subplot,
            color_line=color_line,
            display_factor=display_factor,
            leads=leads,
            length_seconds=length_seconds,
            number_of_leads=number_of_leads,
            rows=rows,
        )

    def _draw_plot_leads(
        self,
        subplot: Subplot,
        color_line: tuple[float, float, float],
        display_factor: float,
        leads: ndarray,
        length_seconds: float,
        number_of_leads: int,
        rows: int,
    ):
        for column in range(self.columns):
            for row in range(rows):
                if column * rows + row < number_of_leads:
                    y_offset = -(self.row_height / 2) * ceil(row % rows)
                    # if y_offset < -5:
                    #     y_offset = y_offset + 0.25

                    x_offset = 0
                    if column > 0:
                        x_offset = length_seconds * column
                        if self.show_separate_line:
                            subplot.plot(
                                [x_offset, x_offset],
                                [
                                    leads[t_lead][0] + y_offset - 0.3,  # noqa
                                    leads[t_lead][0] + y_offset + 0.3,  # noqa
                                ],
                                linewidth=self.line_width * display_factor,
                                color=color_line,
                            )

                    t_lead = self.lead_order[column * rows + row]

                    step = 1.0 / self.sample_rate
                    if self.show_lead_name:
                        subplot.text(
                            x_offset + 0.07,
                            y_offset - 0.5,
                            self.lead_index_order[t_lead],
                            fontsize=9 * display_factor,
                        )
                    subplot.plot(
                        np.arange(0, len(leads[t_lead]) * step, step) + x_offset,
                        leads[t_lead] + y_offset,
                        linewidth=self.line_width * display_factor,
                        color=color_line,
                    )

    def _set_xy_axis_size(self, length_seconds: float, rows: int) -> tuple[float, float, float, float]:
        x_min = 0
        x_max = self.columns * length_seconds
        y_min = self.row_height / 4 - (rows / 2) * self.row_height
        y_max = self.row_height / 4
        return x_max, x_min, y_max, y_min

    @staticmethod
    def _setup_subplot(figure, x_max: float, x_min: float, y_max: float, y_min: float) -> Subplot:
        subplot = figure.add_subplot(1, 1, 1)
        subplot.set_ylim(y_min, y_max)
        subplot.set_xlim(x_min, x_max)

        # remove black line box around plot
        subplot.spines["top"].set_visible(False)
        subplot.spines["right"].set_visible(False)
        subplot.spines["bottom"].set_visible(False)
        subplot.spines["left"].set_visible(False)
        return subplot

    def _setup_figure(self, display_factor: float, rows: int, length_seconds: float) -> Figure:

        figure = plt.figure(
            figsize=(
                length_seconds * self.columns * display_factor,
                rows * self.row_height / 5 * display_factor,
            )
        )
        figure.subplots_adjust(
            hspace=0,
            wspace=0,
            left=0,  # the left side of the subplots of the figure
            right=1,  # the right side of the subplots of the figure
            bottom=0,  # the bottom of the subplots of the figure
            top=1,
        )

        figure.suptitle(self.title)
        if self.adjust_subplots:
            figure.subplots_adjust(bottom=0.00001)
            figure.subplots_adjust(right=0.99999)
            # figure.subplots_adjust(left=0.00001)
            # figure.subplots_adjust(top=0.99999)
        return figure

    def _set_line_style(
        self,
    ) -> tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]:
        if self.color_style == ColorStyle.COLOR:
            color_major_grid = (1, 0, 0)
            color_minor_grid = (1, 0.7, 0.7)
            color_line = (0, 0, 0)
        elif self.color_style in [
            ColorStyle.BLACK_AND_WHITE,
            ColorStyle.GREY_SCALE,
            ColorStyle.MASK,
        ]:
            color_major_grid = (0.4, 0.4, 0.4)
            color_minor_grid = (0.75, 0.75, 0.75)
            color_line = (0, 0, 0)
        else:
            raise ValueError()

        return color_line, color_major_grid, color_minor_grid

        # if self.style == 'bw':
        #     color_major_grid = (0.4, 0.4, 0.4)
        #     color_minor_grid = (0.75, 0.75, 0.75)
        #     color_line = (0, 0, 0)
        # elif self.style == 'tl':
        #     color_major_grid = (0.8, 0.8, 0.8)
        #     color_minor_grid = (0.8, 0.8, 0.8)
        #     color_line = (0, 0, 0.3)
        # elif self.style == 'bl':
        #     color_major_grid = (1, 0, 0)
        #     color_minor_grid = (1, 0.7, 0.7)
        #     color_line = (0, 0, 0.1)
        # else:
        #     color_major_grid = (1, 0, 0)
        #     color_minor_grid = (1, 0.7, 0.7)
        #     color_line = (0, 0, 0.7)
        # return color_line, color_major_grid, color_minor_grid

    @staticmethod
    def _draw_grid(
        subplot: Subplot,
        color_major_grid: tuple[float, float, float],
        color_minor_grid: tuple[float, float, float],
        display_factor: float,
        x_max: float,
        x_min: float,
        y_max: float,
        y_min: float,
    ):
        """

        Args:
            subplot:
            color_major_grid: Line COLOR for major grid lines
            color_minor_grid: Line COLOR for minor grid lines
            display_factor:
            x_max: X axis maximum position value
            x_min: X axis minimum position value
            y_max: Y axis maximum position value
            y_min: Y axis minimum position value

        Returns:

        """
        subplot.set_xticks(np.arange(x_min, x_max, 0.2))
        subplot.set_yticks(np.arange(y_min, y_max, 0.5))
        subplot.minorticks_on()
        subplot.xaxis.set_minor_locator(AutoMinorLocator(5))
        subplot.grid(
            which="major",
            linestyle="-",
            linewidth=0.5 * display_factor,
            color=color_major_grid,
        )
        subplot.grid(
            which="minor",
            linestyle="-",
            linewidth=0.5 * display_factor,
            color=color_minor_grid,
        )

    def _create_plot(self, ecg_leads):

        plt.ioff()
        # plot_start = time.time()

        self._draw_plot(ecg_leads=ecg_leads)
        # plot_finish = time.time()
        # plot_execution_time = plot_finish - plot_start
        # print(f"Plotting took {plot_execution_time:.6f} seconds")
        canvas = plt.get_current_fig_manager().canvas

        plt.close()
        # agg=canvas
        agg = canvas.switch_backends(FigureCanvasAgg)

        # draw_start = time.time()
        agg.draw()
        # draw_finish = time.time()
        # draw_execution_time = draw_finish - draw_start
        # print(f"Drawing took {draw_execution_time:.6f} seconds")
        return np.asarray(agg.buffer_rgba())

    # @staticmethod
    # def _convert_to_mask(image: ndarray) -> ndarray:
    #     return images.convert_to_mask(image)

    def _set_style(self, color_style: ColorStyle, show_grid: bool):
        if show_grid != self.show_grid:
            self.show_grid = show_grid

        if color_style != self.color_style:
            self.color_style = color_style
            if self.color_style in [ColorStyle.MASK]:
                self._set_mask_style()

    def _set_mask_style(self):
        self.show_grid = False
        self.show_lead_name = False
        self.title = ""

    @staticmethod
    def _convert_to_ndarray(ecg_leads: Leads) -> ndarray:
        ecg_leads.leads.sort()
        number_of_leads = len(ecg_leads.leads)
        lead_signal = ecg_leads.leads[0].signal
        number_of_signals = len(lead_signal)
        output_array = np.zeros(shape=[number_of_leads, number_of_signals])
        for i, s in enumerate(ecg_leads.leads):
            p_lead = np.array(s.signal)
            output_array[i] = p_lead
        return output_array

    # convert for protobuf models
    # @staticmethod
    # def _convert_to_ndarray(ecg_leads: Leads) -> ndarray:
    #     ecg_leads.leads.sort()
    #     number_of_leads = len(ecg_leads.leads)
    #     lead_signal = ecg_leads.leads[0].signals
    #     number_of_signals = len(lead_signal)
    #     output_array = np.zeros(shape=[number_of_leads, number_of_signals])
    #     for array_position, lead in enumerate(ecg_leads.leads):
    #         sig = [v.voltage for v in lead.signals]
    #         # g = s.signals
    #         p_lead = np.array(sig)
    #         output_array[array_position] = p_lead
    #     return output_array


def data_directory() -> pathlib.Path:
    file_path = pathlib.Path(pathlib.Path(pathlib.Path(__file__).parent, "grid_images"))
    print(file_path)
    return file_path



import json

# import orjson as orjson
from pydantic import BaseModel
from pydantic.utils import to_camel

#
# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()


class MyBaseModel(BaseModel):
    def __hash__(self):  # make hashable BaseModel subclass
        return hash((type(self),) + tuple(self.__dict__.values()))

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    @classmethod
    def from_json(cls, j):
        return cls.from_dict(json.loads(j))

    def to_json(self):
        return self.json(by_alias=True)

    class Config:
        alias_generator = to_camel
        arbitrary_types_allowed = True
        # use_enum_values = True
        # by_alias = True
        # json_dumps = orjson_dumps



from typing import List

import pydantic

from ecgai_drawing.enums.ecg_lead_name import LeadName
from ecgai_drawing.models.model_base import MyBaseModel


# @dataclass
class Lead(MyBaseModel):
    lead_name: LeadName  # = Field(..., alias='leadName')v
    signal: List[float]  # = Field(..., alias='signal')

    # noinspection PyMethodParameters
    @pydantic.validator("lead_name", pre=True)
    def lead_name_to_enum(cls, v):
        return LeadName[v]

    @classmethod
    def create(cls, lead_name: str, signal: List[float]):
        try:
            d = dict(LeadName=lead_name, Signal=signal)
            return cls.from_dict(d)
        except pydantic.ValidationError as e:
            # logging.error(e)
            raise e

    def __lt__(self, other):
        return self.lead_name < other.lead_name



from numpy import ndarray

from ecgai_drawing import images
from ecgai_drawing.ecg_plot_image import EcgPlotImage
from ecgai_drawing.ecg_plotter import EcgPlotter
from ecgai_drawing.enums.artifact import Artifact
from ecgai_drawing.enums.color_style import ColorStyle
from ecgai_drawing.images import (
    DEFAULT_FILE_EXTENSION,
    convert_from_rgba_to_rgb,
    convert_to_bytes,
)
from ecgai_drawing.models.ecg_leads import Leads


class CreateEcgPlot:
    # _plot_request: EcgPlotRequest
    transaction_id: str
    record_name: str
    sample_rate: int
    color_style: ColorStyle
    show_grid: bool
    artifact: Artifact
    ecg_leads: Leads
    file_name: str

    def __init__(
        self,
        transaction_id: str,
        record_name: str,
        sample_rate: int,
        ecg_leads: Leads,
        file_name: str = "",
        color_style: ColorStyle = ColorStyle.BLACK_AND_WHITE,
        show_grid: bool = True,
        artifact: Artifact = Artifact.NONE,
    ):
        self.transaction_id = transaction_id
        self.record_name = record_name
        self.sample_rate = sample_rate
        self.ecg_leads = ecg_leads
        if file_name == "":
            self.file_name = record_name
        else:
            self.file_name = file_name
        self.color_style = color_style
        self.artifact = artifact
        self.show_grid = show_grid

    # ecg = 6;

    def handle(self) -> EcgPlotImage:
        image = self._create_image()
        return EcgPlotImage.create(
            transaction_id=self.transaction_id,
            record_name=self.record_name,
            file_name=self.file_name,
            file_extension=DEFAULT_FILE_EXTENSION,
            image=convert_to_bytes(image=image),
        )
        # return EcgPlotImage(
        #     transaction_id=self._plot_request.transaction_id,
        #     record_name=self._plot_request.record_name,
        #     file_name=self._plot_request.file_name,
        #     image=convert_to_bytes(image=image),
        # )
        # return draw_response

    # def _create_response(self, image: ndarray) -> EcgPlotRequest:
    #     f, buffer = cv2.imencode(ext='.png', img=image)
    #     image_string = base64.b64encode(buffer)
    #     return EcgPlotRequest(
    #         transaction_id=self.transaction_id,
    #         record_name=self.ecg.record_name,
    #         file_name=self.file_name,
    #         image=image_string,
    #     )
    # return DrawEcgPlotResponseOld.create(transaction_id=self.transaction_id,
    # record_name=self.record_name, file_name="", image=image)

    def _create_image(self) -> ndarray:
        ecg_plotter = EcgPlotter()
        image = ecg_plotter.plot(
            sample_rate=self.sample_rate,
            ecg_leads=self.ecg_leads,
            title=self.record_name,
            color_style=self.color_style,
            show_grid=self.show_grid,
        )
        image = convert_from_rgba_to_rgb(image)
        image = self._add_artifact_to_image(image)
        image = self._convert_format(image)
        self._is_grid_name()
        return image

    def _is_grid_name(self):
        if self.color_style != ColorStyle.MASK:
            if not self.show_grid:
                self.file_name = f"ng_{self.file_name}"

    def _convert_format(self, image):

        if self.color_style in [ColorStyle.COLOR]:
            return image

        if self.color_style == ColorStyle.BLACK_AND_WHITE:
            image = images.convert_to_black_and_white(image=image)
            self.file_name = f"bw_{self.file_name}"
        elif self.color_style == ColorStyle.GREY_SCALE:
            image = images.convert_to_grey_scale(image=image)
            self.file_name = f"gs_{self.file_name}"
        elif self.color_style == ColorStyle.MASK:
            image = images.convert_to_mask(image=image)
            self.file_name = f"mask_{self.file_name}"

            #
            # if self.color_style in [
            #     ColorStyle.MASK,
            # ]:
            #     # image = self.convert_to_binary(image=image)
            #     image = self.convert_to_mask(image=image)

        return image

    def _add_artifact_to_image(self, image):

        if (
            self.color_style != ColorStyle.MASK
            or self.artifact != Artifact.ARTIFACT_UNSPECIFIED
            or self.artifact != Artifact.NONE
        ):

            if self.artifact == Artifact.PEPPER:
                image = images.add_pepper_to_image(image=image)
                self.file_name = f"p_{self.file_name}"
            elif self.artifact == Artifact.SALT:
                image = images.add_salt_to_image(image=image)
                self.file_name = f"s_{self.file_name}"
            elif self.artifact == Artifact.SALT_AND_PEPPER:
                image = images.add_salt_and_pepper_to_image(image=image)
                self.file_name = f"sp_{self.file_name}"
            # else:
            #     raise ValueError()
        return image

    # @staticmethod
    # def convert_to_mask(image: ndarray) -> ndarray:
    #     # return util.invert(image)
    #     return image

    # @staticmethod
    # def convert_to_binary(image: ndarray) -> ndarray:
    #     thresh = threshold_isodata(image)
    #     return image > thresh

    # @staticmethod
    # def add_salt_and_pepper_to_image(image: ndarray) -> ndarray:
    #     return skimage.util.random_noise(image=image, mode="s&p")
    #
    # @staticmethod
    # def add_salt_to_image(image: ndarray) -> ndarray:
    #     return skimage.util.random_noise(image=image, mode="salt")
    #
    # @staticmethod
    # def add_pepper_to_image(image: ndarray) -> ndarray:
    #     return skimage.util.random_noise(image=image, mode="pepper")



from enum import IntEnum


class ColorStyle(IntEnum):
    BLACK_AND_WHITE = 1  # 'BLACK_AND_WHITE'
    COLOR = 2  # 'COLOR'
    MASK = 3  # 'MASK'
    GREY_SCALE = 4
    # binary = 4  # 'binary'

    # @classmethod
    # def _missing_name_(cls, name):
    #     # noinspection PyTypeChecker
    #     for member in cls:
    #         if member.name.lower() == name.lower():
    #             return member
    #
    # def equals(self, string):
    #     return self.name == string



from ecgai_drawing.enums.enum_ordered_base import OrderedEnum


class LeadName(int, OrderedEnum):
    I = 1
    II = 2
    III = 3
    aVR = 4
    aVL = 5
    aVF = 6
    V1 = 7
    V2 = 8
    V3 = 9
    V4 = 10
    V5 = 11
    V6 = 12
    Unknown = 13

    @classmethod
    def _missing_name_(cls, name):
        # noinspection PyTypeChecker
        for member in cls:
            if member.name.lower() == str(name).lower():
                return member
            elif str(member.value) == str(name):
                return member
        return cls.Unknown



from dataclasses import dataclass

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Subplot
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from matplotlib.ticker import AutoMinorLocator

from ecgai_drawing import images
from ecgai_drawing.enums.color_style import ColorStyle


@dataclass
class EcgGridPlotter:
    def __init__(self):
        self.columns = 2
        self.adjust_subplots = True
        self.color_style = ColorStyle.BLACK_AND_WHITE
        self.line_width: float = 0.5
        self.row_height: int = 6
        self.display_factor = 1

    def plot(self, x_length, y_length, color_style):
        # length_seconds = 5000 / self.sample_rate
        # number_of_leads = len(self.lead_order)
        # rows = int(ceil(number_of_leads / self.columns))
        # display_factor = 2.5
        self.color_style = color_style
        # line_width = 0.5
        self._plot(x_length, y_length)

        canvas = plt.get_current_fig_manager().canvas

        plt.close()
        # agg=canvas
        agg = canvas.switch_backends(FigureCanvasAgg)

        agg.draw()
        image = np.asarray(agg.buffer_rgba())
        # convert from rgba format to rgb format
        image = images.convert_from_rgba_to_rgb(image=image)

        # convert from rgb format to bgr format required by openCv
        image = images.convert_from_rgb_to_bgr(image=image)

        return image

    def _plot(self, x_length, y_length):
        figure = self._setup_figure(x_height=x_length, y_length=y_length)
        x_max, x_min, y_max, y_min = self._set_xy_axis_size(10, 6)
        subplot = self._setup_subplot(figure, x_max, x_min, y_max, y_min)
        self.display_factor **= 0.5
        color_line, color_major_grid, color_minor_grid = self._set_line_style()
        self._draw_grid(
            subplot=subplot,
            color_major_grid=color_major_grid,
            color_minor_grid=color_minor_grid,
            display_factor=self.display_factor,
            x_max=x_max,
            x_min=x_min,
            y_max=y_max,
            y_min=y_min,
        )

    def _setup_figure(self, x_height: int, y_length: int) -> Figure:
        x_height = self.pixels_to_inches(x_height)
        y_length = self.pixels_to_inches(y_length)

        figure = plt.figure(
            figsize=(
                x_height,
                y_length,
            )
        )
        figure.subplots_adjust(
            hspace=0,
            wspace=0,
            left=0,  # the left side of the subplots of the figure
            right=1,  # the right side of the subplots of the figure
            bottom=0,  # the bottom of the subplots of the figure
            top=1,
        )

        if self.adjust_subplots:
            figure.subplots_adjust(bottom=0.00001)
            figure.subplots_adjust(right=0.99999)
            # figure.subplots_adjust(left=0.00001)
            # figure.subplots_adjust(top=0.99999)
        return figure

    @staticmethod
    def _draw_grid(
        subplot: Subplot,
        color_major_grid: tuple[float, float, float],
        color_minor_grid: tuple[float, float, float],
        display_factor: float,
        x_max: float,
        x_min: float,
        y_max: float,
        y_min: float,
    ):
        """

        Args:
            subplot:
            color_major_grid: Line COLOR for major grid lines
            color_minor_grid: Line COLOR for minor grid lines
            display_factor:
            x_max: X axis maximum position value
            x_min: X axis minimum position value
            y_max: Y axis maximum position value
            y_min: Y axis minimum position value

        Returns:

        """
        subplot.set_xticks(np.arange(x_min, x_max, 0.2))
        subplot.set_yticks(np.arange(y_min, y_max, 0.5))

        subplot.xaxis.set_minor_locator(AutoMinorLocator(5))

        subplot.grid(
            which="major",
            linestyle="-",
            linewidth=0.5 * display_factor,
            color=color_major_grid,
            zorder=2.5,
        )

        subplot.minorticks_on()

        subplot.grid(
            which="minor",
            linestyle="-",
            linewidth=0.5 * display_factor,
            color=color_minor_grid,
            zorder=0.5,
        )

    @staticmethod
    def _setup_subplot(figure, x_max: float, x_min: float, y_max: float, y_min: float) -> Subplot:
        subplot = figure.add_subplot(1, 1, 1)
        subplot.set_ylim(y_min, y_max)
        subplot.set_xlim(x_min, x_max)

        # remove black line box around plot
        subplot.spines["top"].set_visible(False)
        subplot.spines["right"].set_visible(False)
        subplot.spines["bottom"].set_visible(False)
        subplot.spines["left"].set_visible(False)
        return subplot

    def _set_xy_axis_size(self, length_seconds: float, rows: int) -> tuple[float, float, float, float]:
        x_min = 0
        x_max = self.columns * length_seconds
        y_min = self.row_height / 4 - (rows / 2) * self.row_height
        y_max = self.row_height / 4
        return x_max, x_min, y_max, y_min

    @staticmethod
    def pixels_to_inches(pixels: int, dpi: int = 100) -> float:
        inches = float(pixels / dpi)
        return inches

    def _set_line_style(
        self,
    ) -> tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]:
        if self.color_style == ColorStyle.COLOR:
            color_major_grid = (1, 0, 0)
            color_minor_grid = (1, 0.7, 0.7)
            color_line = (0, 0, 0.7)
        elif self.color_style in [
            ColorStyle.BLACK_AND_WHITE,
            ColorStyle.GREY_SCALE,
            ColorStyle.MASK,
        ]:
            color_major_grid = (0.4, 0.4, 0.4)
            color_minor_grid = (0.75, 0.75, 0.75)
            color_line = (0, 0, 0)
        else:
            raise ValueError()

        return color_line, color_major_grid, color_minor_grid



import base64
import os
from pathlib import Path

import cv2
import numpy as np
import skimage
from numpy import ndarray

DEFAULT_FILE_EXTENSION = ".png"


def load_image(image_name: str, image_path: Path) -> ndarray:
    file_name = os.path.join(image_path, image_name)
    print("the file path for load is " + file_name)
    # return skimage.io.imread(fname=file_name)
    # TODO cv.imread or cv.imwrite is changing the colour of the ecg line from red to blue, find solution
    return cv2.imread(filename=file_name, flags=cv2.IMREAD_UNCHANGED)


def save_image_as_ubyte(image: ndarray, image_name: str, save_path: str):
    byte_image = skimage.img_as_ubyte(image)
    save_image(image=byte_image, image_name=image_name, save_path=save_path)


# def save_image_as_bool(image: ndarray, COLOR_WITHOUT_GRID_NAME: str, save_path: str):
#     byte_image = skimage.img_as_bool(image)
#     save_image(image=byte_image, COLOR_WITHOUT_GRID_NAME=COLOR_WITHOUT_GRID_NAME, save_path=save_path)


def save_image(image: ndarray, image_name: str, save_path: str):
    file_name = os.path.join(save_path, image_name)
    # file_name2 = os.path.join(save_path, image_name+'ski')
    # skimage.io.imsave(fname=file_name, arr=image)
    # image = convert_from_rgb_to_bgr(image)
    cv2.imwrite(filename=file_name, img=image)


def convert_to_mask(image: ndarray) -> ndarray:
    _image = convert_to_black_and_white(image=image)
    return cv2.bitwise_not(_image)
    # return skimage.util.invert(image)


def convert_to_black_and_white(image: ndarray) -> ndarray:
    _image = convert_to_grey_scale(image=image)
    (_, _image) = cv2.threshold(_image, 150, 255, cv2.THRESH_BINARY)
    return _image
    # thresh = skimage.filters.threshold_isodata(image)
    # return image > thresh


def convert_to_grey_scale(image: ndarray) -> ndarray:
    return cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    # return skimage.COLOR.rgb2gray(image)


def convert_to_bytes(image: ndarray) -> bytes:
    return cv2.imencode(DEFAULT_FILE_EXTENSION, image)[1].tobytes()


def convert_from_bytes(stream) -> ndarray:
    image = np.frombuffer(stream, dtype=np.uint8)  # <class 'numpy.ndarray'>
    # image = np.fromstring(stream, dtype=np.uint8)

    return cv2.imdecode(image, cv2.IMREAD_UNCHANGED)


def convert_to_base64_string(image: ndarray) -> str:
    image_array = cv2.imencode(DEFAULT_FILE_EXTENSION, image)[1]
    return base64.b64encode(image_array).decode()


def convert_from_base64_string(stream: str) -> ndarray:
    jpg_original = base64.b64decode(stream)
    jpg_as_np = np.frombuffer(jpg_original, dtype=np.uint8)
    img = cv2.imdecode(jpg_as_np, flags=1)
    return img


#
# def combine_color_images(base_image: ndarray, top_image: ndarray) -> ndarray:
#     rows, cols, channels = top_image.shape
#     roi = base_image[0:rows, 0:cols]
#
#     top_image_gray = convert_to_grey_scale(top_image)
#
#     ret, mask = cv2.threshold(top_image_gray, 200, 255, cv2.THRESH_BINARY)
#
#     mask_inv = cv2.bitwise_not(mask)
#
#     img_back = cv2.bitwise_and(roi, roi, mask=mask)
#
#     top_image_colour = cv2.bitwise_and(top_image, top_image, mask=mask_inv)
#     # return top_image_colour
#     img_out = cv2.add(img_back, top_image_colour)
#     base_image[0:rows, 0:cols] = img_out
#     return base_image
#
#
# def combine_grey_scale_images(base_image: ndarray, top_image: ndarray) -> ndarray:
#     top_rows, top_cols, top_channels = top_image.shape
#     base_rows, base_cols, base_channels = base_image.shape
#
#     start_row = int((base_rows - top_rows) / 2)
#     start_col = int((base_cols - top_cols) / 2)
#     white = [255, 255, 255]
#     top_image_border = cv2.copyMakeBorder(top_image,
#                                           start_row,
#                                           start_row,
#                                           start_col,
#                                           start_col,
#                                           cv2.BORDER_CONSTANT,
#                                           value=white)
#     border_rows, border_cols, border_channels = top_image_border.shape
#
#     roi = base_image[0:base_rows, 0:base_cols]
#
#     top_image_gray = convert_to_grey_scale(top_image_border)
#     base_image_gray = convert_to_grey_scale(base_image)
#     ret, mask = cv2.threshold(top_image_gray, 200, 255, cv2.THRESH_BINARY)
#
#     mask_inv = cv2.bitwise_not(mask)
#
#     img_back = cv2.bitwise_and(roi, roi, mask=mask)
#
#     top_image_colour = cv2.bitwise_and(top_image_border, top_image_border, mask=mask_inv)
#     # return top_image_colour
#     # top_image_out = convert_to_grey_scale(top_image_colour)
#     img_out = cv2.add(img_back, top_image_colour)
#     base_image[0:border_rows, 0:border_cols] = img_out
#     return base_image


def combine_images(base_image: ndarray, top_image: ndarray) -> ndarray:
    top_rows, top_cols, top_channels = top_image.shape
    base_rows, base_cols, base_channels = base_image.shape

    start_row = int((base_rows - top_rows) / 2)
    start_col = int((base_cols - top_cols) / 2)

    if start_row < 0 or start_col < 0:
        raise ValueError("Top image is too big for base image")

    white = [255, 255, 255]
    top_image_border = cv2.copyMakeBorder(
        top_image, start_row, start_row, start_col, start_col, cv2.BORDER_CONSTANT, value=white
    )
    border_rows, border_cols, border_channels = top_image_border.shape

    roi = base_image[0:base_rows, 0:base_cols]

    top_image_gray = convert_to_grey_scale(top_image_border)
    # base_image_gray = convert_to_grey_scale(base_image)
    ret, mask = cv2.threshold(top_image_gray, 200, 255, cv2.THRESH_BINARY)

    mask_inv = cv2.bitwise_not(mask)

    img_back = cv2.bitwise_and(roi, roi, mask=mask)

    top_image_colour = cv2.bitwise_and(top_image_border, top_image_border, mask=mask_inv)
    # return top_image_colour
    # top_image_out = convert_to_grey_scale(top_image_colour)
    img_out = cv2.add(img_back, top_image_colour)
    base_image[0:border_rows, 0:border_cols] = img_out
    return base_image


# def create_binary_image(COLOR_WITHOUT_GRID_NAME: str, image_path: str) -> ndarray:
#     image = load_image(COLOR_WITHOUT_GRID_NAME=COLOR_WITHOUT_GRID_NAME, image_path=image_path, as_gray=True)
#     thresh = threshold_isodata(image)
#     return image > thresh

# def create_binary_image(image: ndarray) -> ndarray:
#     thresh = threshold_isodata(image)
#     return image > thresh
#
#
# def convert_to_grey_scale(image: ndarray) -> ndarray:
#     image = rgba2rgb(image)
#     return rgb2gray(image)
#
#
def convert_from_rgba_to_rgb(image: ndarray) -> ndarray:
    return cv2.cvtColor(image, cv2.COLOR_RGBA2RGB)
    # return skimage.COLOR.rgba2rgb(image)


def convert_from_brg_to_rgb(image: ndarray) -> ndarray:
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)


def convert_from_rgb_to_bgr(image: ndarray) -> ndarray:
    return cv2.cvtColor(image, cv2.COLOR_RGB2BGR)


#

#
#
# #
# # def create_bool_image(image: ndarray):
# #     bool_image = img_as_bool(image=image)
# #     return bool_image
#
#
# def invert_image(image: ndarray) -> ndarray:
#     return util.invert(image)
#
#

# def add_salt_and_pepper_to_image(image: ndarray):
#     prob = 0.05
#     output = np.zeros(image.shape, np.uint8)
#     thres = 1 - prob
#     for i in range(image.shape[0]):
#         for j in range(image.shape[1]):
#             rdn = random.random()
#             if rdn < prob:
#                 output[i][j] = 0
#             elif rdn > thres:
#                 output[i][j] = 255
#             else:
#                 output[i][j] = image[i][j]
#     return output


def add_salt_and_pepper_to_image(image: ndarray, amount: float = 0.05, salt_vs_pepper: float = 0.5) -> ndarray:
    """
    Function to add random salt and pepper noise to a ndarray image.

    Original source https://gist.github.com/gutierrezps/f4ddad3bbd2ad5a9b96e3c06378e28b4
    Parameters
    ----------
    image: ndarray
        Input image data.
    amount: float, optional
        Proportion of image pixels to replace with noise on range [0, 1].
        Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05
    salt_vs_pepper: float, optional
        Proportion of salt vs. pepper noise for 's&p' on range [0, 1].
        Higher values represent more salt. Default : 0.5 (equal amounts)

    Returns
    -------
    out : ndarray
        Output floating-point image data on range [0, 1] or [-1, 1] if the
        input `image` was unsigned or signed, respectively.
    """
    if 1.0 < salt_vs_pepper < 0.0:
        raise ValueError("salt_vs_pepper value needs to be equal or between 0.0 - 1.0")

    out = image.copy()
    if len(image.shape) == 2:
        black = 0
        white = 255
    else:
        colorspace = image.shape[2]
        if colorspace == 3:  # RGB
            black = np.array([0, 0, 0], dtype="uint8")
            white = np.array([255, 255, 255], dtype="uint8")
        else:  # RGBA
            black = np.array([0, 0, 0, 255], dtype="uint8")
            white = np.array([255, 255, 255, 255], dtype="uint8")
    probs = np.random.random(out.shape[:2])

    pepper_vs_salt = 1.0 - salt_vs_pepper

    out[probs < (amount * pepper_vs_salt)] = black
    out[probs > 1 - (amount * salt_vs_pepper)] = white
    return out

    # return sp_noise(image=image, amount=0.1)


#     # Getting the dimensions of the image
#     row, col, _ = image.shape
#     number_of_pixels = int((row * col) * 0.5)
#     # Randomly pick some pixels in the
#     # image for coloring them white
#     # Pick a random number between 300 and 10000
#     # number_of_pixels = random.randint(90000, 100000)
#     for i in range(number_of_pixels):
#         # Pick a random y coordinate
#         y_coord = random.randint(0, row - 1)
#
#         # Pick a random x coordinate
#         x_coord = random.randint(0, col - 1)
#
#         # Color that pixel to white
#         image[y_coord][x_coord] = 255
#
#     # Randomly pick some pixels in
#     # the image for coloring them black
#     # Pick a random number between 300 and 10000
#     number_of_pixels = random.randint(300, 10000)
#     for i in range(number_of_pixels):
#         # Pick a random y coordinate
#         y_coord = random.randint(0, row - 1)
#
#         # Pick a random x coordinate
#         x_coord = random.randint(0, col - 1)
#
#         # Color that pixel to black
#         image[y_coord][x_coord] = 0
#
#     return image


def add_salt_to_image(image: ndarray, amount: float = 0.05) -> ndarray:
    """
    Function to add random salt noise to a ndarray image.

    Original source https://gist.github.com/gutierrezps/f4ddad3bbd2ad5a9b96e3c06378e28b4
    Parameters
    ----------
    image: ndarray
        Input image data.
    amount: float, optional
        Proportion of image pixels to replace with noise on range [0, 1].
        Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05

    Returns
    -------
    out : ndarray
        Output floating-point image data on range [0, 1] or [-1, 1] if the
        input `image` was unsigned or signed, respectively.
    """
    return add_salt_and_pepper_to_image(image=image, amount=amount, salt_vs_pepper=1.0)


def add_pepper_to_image(image: ndarray, amount: float = 0.05) -> ndarray:
    """
    Function to add random pepper noise to a ndarray image.

    Original source https://gist.github.com/gutierrezps/f4ddad3bbd2ad5a9b96e3c06378e28b4
    Parameters
    ----------
    image: ndarray
        Input image data.
    amount: float, optional
        Proportion of image pixels to replace with noise on range [0, 1].
        Used in 'salt', 'pepper', and 'salt & pepper'. Default : 0.05

    Returns
    -------
    out : ndarray
        Output floating-point image data on range [0, 1] or [-1, 1] if the
        input `image` was unsigned or signed, respectively.
    """
    return add_salt_and_pepper_to_image(image=image, amount=amount, salt_vs_pepper=0.0)



from typing import List

from ecgai_drawing.models.ecg_lead import Lead
from ecgai_drawing.models.model_base import MyBaseModel


class Leads(MyBaseModel):
    record_name: str
    sample_rate: int
    leads: List[Lead]

    @classmethod
    def create(cls, record_name: str, sample_rate: int, leads: List[Lead]):
        d = dict(RecordName=record_name, SampleRate=sample_rate, Leads=leads)
        return cls.from_dict(d)

    # def to_json(self):
    #
    #     """
    #     Not implemented in this model as issues with EcgLeadNameOld not Json serializable and this model will not be saved
    #     """
    #     raise NotImplementedError()
